#!/usr/bin/env python
# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import os
import shutil
import subprocess
import sys
import textwrap

from charmhelpers import core
from charmhelpers import fetch


def juju_info(msg, level='INFO'):
    core.hookenv.log(msg, level)


def _service_name(config):
    unit = core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return unit


def _service_dir(config):
    unit = core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(config['install_root'], unit)


def _upstart_log_dir(config):
    return os.path.join(_service_dir(config), 'logs')


def _lander_logs_dir(config):
    return os.path.join(_service_dir(config), 'lander-logs')


def _relation_file(config):
    relation = os.environ['JUJU_RELATION_ID']
    return os.path.join(_service_dir(config), 'juju-relations', relation)


def add_dependencies(config):
    juju_info('Adding dependencies.')
    sources = config['sources'].split(';')
    for source in sources:
        fetch.add_source(source)
    fetch.apt_update()
    for package in config['packages'].split():
        fetch.apt_install(package, fatal=True)


def _install_from_tarball(config, upgrade):
    juju_info('grabbing service from tarball...')
    sdir = _service_dir(config)
    if not upgrade:
        if os.path.exists(sdir):
            juju_info('deleting pre-existing service directory: %s' % sdir)
            shutil.rmtree(sdir)
        os.mkdir(sdir)
    cmd = 'curl %s | tar -xzC %s' % (config['tarball'], sdir)
    subprocess.check_call(cmd, shell=True)


def _install_from_bzr(config, upgrade):
    juju_info('grabbing service from bzr...')

    sdir = _service_dir(config)
    rev = config.get('revno', '')

    if upgrade:
        args = ['bzr', 'pull']
        if rev:
            args.extend(['-r', rev])
        subprocess.check_call(args, cwd=sdir)
        return

    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)

    args = ['bzr', 'branch']
    if rev:
        args.extend(['-r', rev])
    args.append(config['branch'])
    args.append(sdir)
    subprocess.check_call(args)


def install(config, upgrade=False):
    if config.get('vcs') == 'tarball':
        _install_from_tarball(config, upgrade)
    else:
        _install_from_bzr(config, upgrade)


def upgrade_charm(config):
    install(config, True)
    restart_upstart(config)


def create_unit_config(config):
    unit_config = config.get('unit-config')
    if unit_config:
        with open(os.path.join(_service_dir(config), 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


def create_upstart(config):
    template = textwrap.dedent('''
    #--------------------------------------------------------------
    # This file is managed by the lander Juju charm
    #--------------------------------------------------------------
    description "starts a lander worker"
    start on (local-filesystems and net-device-up IFACE=eth0)
    stop on runlevel [!12345]

    # If the process quits unexpectedly trigger respawn it
    respawn
    # unless it fails 15 times within 5 seconds
    respawn limit 15 5

    chdir {sdir}

    # setuid/setgid don't work because we need to do mkdir as root
    # this is the easiest work-around:
    script
      [ -d {logdir} ] || mkdir {logdir}
      [ -d {lander_logs} ] || mkdir {lander_logs}
      chown {uid}:{gid} {logdir} {lander_logs}
      chmod 755 {logdir} {lander_logs}
      exec su -c "{main}" {uid} >>{logdir}/upstart.log 2>&1
    end script
    ''')
    params = {
        'main': config['main'],
        'sdir': _service_dir(config),
        'uid': config.get('uid', 'nobody'),
        'gid': config.get('gid', 'nogroup'),
        'logdir': _upstart_log_dir(config),
        'lander_logs': _lander_logs_dir(config)
    }
    with open('/etc/init/%s.conf' % _service_name(config), 'w') as f:
        f.write(template.format(**params))
        # need this so the file is ready for the subprocess call
        f.flush()
        os.fsync(f.fileno())


def restart_upstart(config):
    try:
        subprocess.check_call(['/sbin/restart', _service_name(config)])
    except:
        # if it wasn't running, restart fails, so just try to start
        subprocess.check_call(['/sbin/start', _service_name(config)])


def config_changed(config):
    juju_info("Config: %s" % config)
    add_dependencies(config)
    create_unit_config(config)
    create_upstart(config)
    restart_upstart(config)


def lander_relation_joined(config):
    rfile = _relation_file(config)
    rdir = os.path.dirname(rfile)
    if not os.path.exists(rdir):
        os.mkdir(rdir)
    with open(rfile, 'w') as f:
        unit = core.hookenv.remote_unit().split('/')[0]
        addr = core.hookenv.relation_get('private-address')
        f.write('%s: %s' % (unit, addr))


def lander_relation_changed(config):
    lander_relation_joined(config)


def lander_relation_broken(config):
    rfile = _relation_file(config)
    juju_info('removing relation file: %s' % rfile)
    if os.path.exists(rfile):
        os.unlink(rfile)


def amqp_relation_joined(config):
    core.hookenv.relation_set(None, {
        'username': config['amqp-user'],
        'vhost': config['amqp-vhost'],
    })


def amqp_relation_changed(config):
    addr = core.hookenv.relation_get('private-address')
    password = core.hookenv.relation_get('password')
    with open(os.path.join(_service_dir(config), 'amqp_config.py'), 'w') as f:
        f.write('# DO NOT EDIT. Generated by lander charm hook\n')
        f.write('AMQP_USER = "%s"\n' % config['amqp-user'])
        f.write('AMQP_VHOST = "%s"\n' % config['amqp-vhost'])
        f.write('AMQP_HOST = "%s"\n' % addr)
        f.write('AMQP_PASSWORD = "%s"\n' % password)


def amqp_relation_broken(config):
    os.unlink(os.path.join(_service_dir(config), 'amqp_config.py'))


def json_status_relation_joined(config):
    host = core.hookenv.unit_get('public-address')
    if not host:
        juju_info('no public address found, using private')
        host = core.hookenv.unit_private_ip()
    url = 'http://%s/json_status' % host
    core.hookenv.relation_set(None, {'status-url': url})


def main():
    hook = os.path.basename(sys.argv[0])
    juju_info("Running hook: %s" % hook)

    hook_py = hook.replace('-', '_')
    funcs = globals()
    if hook_py not in funcs:
        print("Unknown hook: %s" % hook)
        return 1

    config = core.hookenv.config()
    try:
        return funcs[hook_py](config)
    except subprocess.CalledProcessError as e:
        juju_info('Error running: %s: %s' % (e.cmd, e.output))
        return e.returncode


if __name__ == '__main__':
    sys.exit(main())
