#!/usr/bin/env python

import base64
import json
import os
import shutil
import subprocess
import sys
import time

from charmhelpers.contrib.charmsupport.nrpe import NRPE

import charmhelpers.core.hookenv
import charmhelpers.fetch
import charmhelpers.payload.execd

hooks = charmhelpers.core.hookenv.Hooks()
_service_name = 'restish'


def juju_info(msg):
    charmhelpers.core.hookenv.log(msg, 'INFO')


def _service_dir(config):
    unit = charmhelpers.core.hookenv.service_name()
    for x in (':', '/', '"', "'"):
        unit = unit.replace(x, '_')
    return os.path.join(config['install_root'], unit)


def _install_from_tarball(config, upgrade):
    juju_info('grabbing service from tarball...')
    sdir = _service_dir(config)
    if not upgrade:
        if os.path.exists(sdir):
            juju_info('deleting pre-existing service directory: %s' % sdir)
            shutil.rmtree(sdir)
        os.mkdir(sdir)
    cmd = 'curl %s | tar -xzC %s' % (config['tarball'], sdir)
    subprocess.check_call(cmd, shell=True)


def _install_from_bzr(config, upgrade):
    juju_info('grabbing service from bzr...')

    sdir = _service_dir(config)
    rev = config.get('revno', '')

    if upgrade:
        args = ['bzr', 'pull']
        if rev:
            args.extend(['-r', rev])
        subprocess.check_call(args, cwd=sdir)
        return

    if os.path.exists(sdir):
        juju_info('deleting pre-existing service directory: %s' % sdir)
        shutil.rmtree(sdir)

    args = ['bzr', 'branch']
    if rev:
        args.extend(['-r', rev])
    args.append(config['branch'])
    args.append(sdir)
    subprocess.check_call(args)


@hooks.hook()
def install(upgrade=False):
    # Support Canonistack/Prodstack basenode subordinate charm interface.
    charmhelpers.payload.execd.execd_preinstall()
    config = charmhelpers.core.hookenv.config()

    pkgs = [x for x in config['packages'].split(' ') if x]
    pkgs.append('bzr')

    framework = config['framework']
    if framework == 'restish':
        pkgs.append('python-restish')
    elif framework == 'pyramid':
        pkgs.append('python-pyramid')
    else:
        juju_info('Invalid framework specified: {}'.format(framework))
        sys.exit(1)

    juju_info('installing apt packages...')
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(pkgs, fatal=True)

    if config.get('vcs') == 'tarball':
        _install_from_tarball(config, upgrade)
    else:
        _install_from_bzr(config, upgrade)

    unit_config = config.get('unit-config')
    if unit_config is not None:
        root_dir = _service_dir(config)
        with open(os.path.join(root_dir, 'unit_config'), 'w') as f:
            f.write(base64.b64decode(unit_config))


@hooks.hook()
def upgrade_charm():
    install(True)
    start()
    _db_migrate()


@hooks.hook()
def start():
    conf = '/etc/init/%s.sh' % _service_name
    if os.path.exists(conf):
        try:
            subprocess.check_call(['service', 'restart', _service_name])
        except subprocess.CalledProcessError:
            # if it wasn't running, restart fails, so just try to start
            subprocess.check_call(['service', 'start', _service_name])


@hooks.hook()
def stop():
    conf = '/etc/init/%s.sh' % _service_name
    if os.path.exists(conf):
        subprocess.check_call(['service', 'stop', _service_name])


def _wsgi_reload():
    # Trigger WSGI reloading
    for relid in charmhelpers.core.hookenv.relation_ids():
        charmhelpers.core.hookenv.relation_set(
            relid, {'wsgi_timestamp': time.time()})


@hooks.hook()
def config_changed():
    # Trigger NRPE config update.
    update_nrpe_config()
    _wsgi_reload()


@hooks.hook('website-relation-changed')
@hooks.hook()
def website_relation_joined():
    config = charmhelpers.core.hookenv.config()
    host = subprocess.check_output(['unit-get', 'private-address']).strip()
    charmhelpers.core.hookenv.relation_set(
        None, {'port': config["port"], 'hostname': host})


@hooks.hook('json_status-relation-joined')
def json_status_relation_joined():
    config = charmhelpers.core.hookenv.config()
    path = config.get('json_status_path')
    if path:
        port = config['port']
        host = charmhelpers.core.hookenv.unit_get('public-address')
        if not host:
            juju_info('no public address found, using private')
            host = charmhelpers.core.hookenv.unit_private_ip()
        url = 'http://%s:%s/%s' % (host, port, path)
        charmhelpers.core.hookenv.relation_set(None, {'status-url': url})


@hooks.hook('wsgi-relation-changed')
@hooks.hook()
def wsgi_relation_joined():
    config = charmhelpers.core.hookenv.config()
    sdir = _service_dir(config)
    ldir = os.path.join(sdir, 'logs')

    if not os.path.exists(ldir):
        charmhelpers.core.host.mkdir(ldir, 'www-data', 'www-data', 0755)

    full_path = []
    path = config.get('python_path')
    if path:
        full_path = []
        for p in path.split(':'):
            if p[0] != '/':
                p = os.path.join(sdir, p)
            full_path.append(p)

    charmhelpers.core.hookenv.relation_set(relation_settings={
        'working_dir': sdir,
        'wsgi_log_file': os.path.join(ldir, 'gunicorn.log'),
        'wsgi_access_logfile': os.path.join(ldir, 'access.log'),
        'port': config['port'],
        'python_path': ':'.join(full_path),
    })

    charmhelpers.core.hookenv.open_port(config['port'])


@hooks.hook()
def wsgi_relation_broken():
    port = charmhelpers.core.hookenv.config()['port']
    charmhelpers.core.hookenv.close_port(port)


@hooks.hook('nrpe-external-master-relation-changed')
@hooks.hook()
def nrpe_external_master_relation_joined():
    update_nrpe_config()


@hooks.hook('local-monitors-relation-changed')
@hooks.hook()
def local_monitors_relation_joined():
    update_nrpe_config()


def update_nrpe_config():
    nrpe_compat = NRPE()
    conf = nrpe_compat.config
    check_http_params = conf.get('nagios_check_http_params')
    if check_http_params:
        nrpe_compat.add_check(
            shortname='restish',
            description='Check Restish Server',
            check_cmd='check_http %s' % check_http_params
        )
    nrpe_compat.write()


def _db_migrate():
    config = charmhelpers.core.hookenv.config()
    cmd = config.get('db_migration_cmd')
    if cmd:
        subprocess.check_call([cmd], shell=True, cwd=_service_dir(config))


def _pgsql_file(config):
    return os.path.join(_service_dir(config), 'pgsql.json')


@hooks.hook('pgsql-relation-joined')
@hooks.hook('pgsql-relation-changed')
def pgsql_relation_joined_changed():
    config = charmhelpers.core.hookenv.config()
    packages = ["python-psycopg2", "postgresql-client"]
    charmhelpers.fetch.apt_install(packages, fatal=True)

    relation = charmhelpers.core.hookenv.relation_get()
    with open(_pgsql_file(config), 'w') as f:
        json.dump(relation, f)

    # some relation calls will just have "private-address". We need to ensure
    # we've gotten all the db credentials before trying to migrate
    if relation.get('user') and relation.get('host'):
        _db_migrate()
    _wsgi_reload()


@hooks.hook()
def pgsql_relation_broken():
    config = charmhelpers.core.hookenv.config()
    settings = _pgsql_file(config)
    if os.path.exists(settings):
        os.unlink(settings)

    _wsgi_reload()


if __name__ == "__main__":
    hooks.execute(sys.argv)
