# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import mock
import os
import unittest
import shutil
import tempfile

import yaml

from intercom import hooks as intercom_hooks
from hooks import hooks


HTTP_REL = {
    'hostname': 'remote.com',
    'port': '8080',
}


INFO_REL = {
    'private-address': '10.0.0.1',
}


PATCHES = [
    'charmhelpers.core.hookenv.close_port',
    'charmhelpers.core.hookenv.config',
    'charmhelpers.core.hookenv.local_unit',
    'charmhelpers.core.hookenv.log',
    'charmhelpers.core.hookenv.open_port',
    'charmhelpers.core.hookenv.remote_unit',
    'charmhelpers.core.hookenv.relation_get',
    'charmhelpers.core.hookenv.relation_id',
    'charmhelpers.core.hookenv.relation_set',
    'charmhelpers.core.hookenv.unit_get',
    'charmhelpers.core.host.mkdir',
    'charmhelpers.fetch.apt_install',
    'charmhelpers.fetch.configure_sources',
    'charmhelpers.payload.execd.execd_preinstall',
    'intercom.log',
    'intercom.get_service_dir',
    'os.environ',
]


class RestishTestCase(unittest.TestCase):
    def setUp(self):
        """Setup a testing hook environment.

        Patches the required charmhelper and intercom objects and
        initializes few mocks for tests.
        """
        super(RestishTestCase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)

        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)

        self.local_unit.return_value = 'local-test'
        self.remote_unit.return_value = 'remote-test'
        self.relation_get.side_effect = (
            lambda rel_name: HTTP_REL.get(rel_name))
        self.relation_id.return_value = '1'
        self.get_service_dir.return_value = self.tmpdir

        def env_get(name):
            return {
                'JUJU_UNIT_NAME': 'restish/0',
            }[name]
        self.environ.__getitem__.side_effect = env_get

        self.unit_get.return_value = '127.0.0.1'

        self._set_default_config()

    def _set_default_config(self):
        defaults = {}
        config = os.path.join(os.path.dirname(__file__), '../config.yaml')
        if os.path.exists(config):
            with open(config) as f:
                config = yaml.safe_load(f)
                for name, option in config['options'].iteritems():
                    defaults[name] = option.get('default')
        self.config.return_value = defaults


class TestInstallHook(RestishTestCase):

    @mock.patch('hooks._install_from_bzr')
    def test_framework_restish(self, install_from_bzr):
        hooks.execute(['hooks/install'])
        pkgs = self.apt_install.call_args_list[0][0][0]
        self.assertIn('python-restish', pkgs)

    @mock.patch('hooks._install_from_bzr')
    def test_framework_pyramid(self, install_from_bzr):
        self.config.return_value['framework'] = 'pyramid'
        hooks.execute(['hooks/install'])
        pkgs = self.apt_install.call_args_list[0][0][0]
        self.assertIn('python-pyramid', pkgs)


class TestIntercomHooks(RestishTestCase):

    def get_intercom_content(self):
        """Utility for getting the stored relationships."""
        content = open(os.path.join(self.tmpdir, 'intercom.json')).read()
        return json.loads(content)

    def set_intercom_content(self, registry):
        """Utility for populating relationships."""
        content = json.dumps(registry)
        with open(os.path.join(self.tmpdir, 'intercom.json'), 'w') as fd:
            fd.write(content)

    def test_intercom_changed_http(self):
        # New 'intercom' relations from interface 'http' are
        # stored in the JSON registry as <hostname>:<port>
        intercom_hooks.execute(['hooks/intercom-relation-changed'])
        self.log.assert_called_with(
            "Intercom unit 'remote-test' (remote.com) joined.")
        self.assertEquals(
            {'1': {'hostname': 'remote.com',
                   'port': '8080',
                   'unit': 'remote-test'}},
            self.get_intercom_content())

    def test_intercom_changed_info(self):
        # New 'intercom' relations from interface 'juju-info' are
        # stored in the JSON registry as <private-address>:80
        self.relation_get.side_effect = (
            lambda rel_name: INFO_REL.get(rel_name))
        intercom_hooks.execute(['hooks/intercom-relation-changed'])
        self.log.assert_called_with(
            "Intercom unit 'remote-test' (10.0.0.1) joined.")
        self.assertEquals(
            {'1': {'hostname': '10.0.0.1',
                   'port': '80',
                   'unit': 'remote-test'}},
            self.get_intercom_content())

    def test_intercom_broken(self):
        # Broken 'intercom' relations are removed from the JSON registry.
        # 'remote_unit' and any relation are not available for 'broken'.
        self.remote_unit.return_value = ''
        self.relation_get.side_effect = lambda rel_name: ''
        # Create an initial intercom.json content.
        self.set_intercom_content(
            {'1': 'to be gone',
             '2': {'hostname': '10.0.0.2',
                   'port': '9000',
                   'unit': 'preserved'}})
        intercom_hooks.execute(['hooks/intercom-relation-broken'])
        self.log.assert_called_with("Intercom unit '' () departed.")
        self.assertEquals(
            {'2': {'hostname': '10.0.0.2',
                   'port': '9000',
                   'unit': 'preserved'}},
            self.get_intercom_content())


class TestWSGIHooks(RestishTestCase):

    def test_wsgi_joined(self):
        hooks.execute(['hooks/wsgi-relation-joined'])
        self.open_port.assert_called_once_with(8080)
        expected = {
            'python_path': '',
            'working_dir': '/srv/local-test',
            'port': 8080,
            'wsgi_access_logfile': '/srv/local-test/logs/access.log',
            'wsgi_log_file': '/srv/local-test/logs/gunicorn.log',
        }
        settings = self.relation_set.call_args_list[0][1]['relation_settings']
        self.assertEqual(expected, settings)
        self.mkdir.assert_called_once_with(
            '/srv/local-test/logs', 'www-data', 'www-data', 0755)

    def test_wsgi_broken(self):
        hooks.execute(['hooks/wsgi-relation-broken'])
        self.close_port.assert_called_once_with(8080)


class TestJSONStatusHooks(RestishTestCase):

    def test_json_status_joined(self):
        self.config.return_value['json_status_path'] = 'foo'
        hooks.execute(['hooks/json_status-relation-joined'])
        self.assertIn('status-url', self.relation_set.call_args_list[0][0][1])


class TestPGSQLHooks(RestishTestCase):
    def setUp(self):
        super(TestPGSQLHooks, self).setUp()
        m = mock.patch('hooks._service_dir')
        m.start().return_value = self.tmpdir
        self.addCleanup(m.stop)

        m = mock.patch('hooks._wsgi_reload')
        self.wsgi_reload = m.start()
        self.addCleanup(m.stop)

        self.relation_get.side_effect = [{'user': 'data', 'host': 'h'}]

    def test_pgsql_joined(self):
        hooks.execute(['hooks/pgsql-relation-joined'])
        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, 'pgsql.json')))
        self.assertEqual(1, self.wsgi_reload.call_count)

    def _do_migration(self, hook):
        with open(os.path.join(self.tmpdir, 'tmp.sh'), 'w') as f:
            f.write('rm $0   # just delete ourself')
            f.flush()
        self.config.return_value['db_migration_cmd'] = 'dash ./tmp.sh'
        hooks.execute([hook])
        self.assertFalse(os.path.exists(os.path.join(self.tmpdir, 'tmp.sh')))

    def test_pgsql_joined_migration(self):
        self._do_migration('hooks/pgsql-relation-joined')
        self.assertEqual(1, self.wsgi_reload.call_count)

    @mock.patch('hooks._install_from_bzr')
    def test_migration_on_upgrade(self, install_from_bzr):
        self._do_migration('hooks/upgrade-charm')

    def test_pgsql_broken(self):
        hooks.execute(['hooks/pgsql-relation-broken'])
        self.assertEqual(1, self.wsgi_reload.call_count)
