# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import subprocess

import swiftclient
import yaml

from novaclient import exceptions as nova_exceptions
from novaclient.v1_1 import client as nova_client

from ucitests import features

# Provide a shorcut so users don't have to import from ucitests
requires = features.requires


class Swift(features.Feature):

    def _probe(self):
        conn = self.get_connection()
        if conn is None:
            return False
        try:
            conn.head_account()
        except swiftclient.exceptions.ClientException:
            return False
        return True

    def get_connection(self):
        try:
            return swiftclient.client.Connection(
                authurl=os.environ['OS_AUTH_URL'],
                user=os.environ['OS_USERNAME'],
                key=os.environ['OS_PASSWORD'],
                os_options={'tenant_name': os.environ['OS_TENANT_NAME'],
                            'region_name': os.environ['OS_REGION_NAME']},
                auth_version='2.0')
        except KeyError:
            # If we miss some env vars, we can't connect
            return None


# The single instance shared by all tests
swift = Swift()


class NovaCompute(features.Feature):

    def _probe(self):
        client = self.get_client()
        if client is None:
            return False
        try:
            client.authenticate()
        except nova_exceptions.ClientException:
            return False
        return True

    def get_client(self):
        try:
            return nova_client.Client(
                os.environ['OS_USERNAME'],
                os.environ['OS_PASSWORD'],
                project_id=os.environ['OS_TENANT_NAME'],
                auth_url=os.environ['OS_AUTH_URL'],
                region_name=os.environ['OS_REGION_NAME'],
                service_type='compute')
        except KeyError:
            # If we miss some env vars, we can't get a client
            return None


# The single instance shared by all tests
nova_compute = NovaCompute()


class BootstrappedJuju(features.Feature):

    def __init__(self):
        super(BootstrappedJuju, self).__init__()
        self.envs = None
        self.cur_env = None
        self.swift = swift
        self.nova = nova_compute

    def _probe_lxc(self, env):
        # We can connect to swift
        if not self.swift.available():
            return False

        try:
            subprocess.check_output(['juju', 'status'], stderr=subprocess.PIPE)
        except:
            return False

        return True

    def _probe_cloud(self, env):
        # There is a bucket
        control_bucket_id = self.get_control_bucket_id()
        if control_bucket_id is None:
            return False
        # We can connect to swift
        if not self.swift.available():
            return False
        # juju put a provider-state file in its control bucket
        provider_state = self.get_provider_state()
        if provider_state is None:
            return False
        # which contains the nova id of the state server node
        nova_id = self.get_nova_id(provider_state)
        if nova_id is None:
            return False
        if not self.nova.available():
            return False
        # if it's still alive, get the nova instance
        instance = self.get_nova_instance(nova_id)
        if instance is None:
            return False
        nova_ip = self.get_nova_ip(instance)
        if nova_ip is None:
            return False
        if not self.is_ip_private(nova_ip):
            # the state server is reachable, a bootstrapped juju is available
            return True
        # A tunnel is needed
        sshuttle_pid = self.get_pid_of_sshuttle(nova_ip)
        if sshuttle_pid is None:
            return False
        return True

    def _probe(self):
        # There is a defined juju env
        env = self.get_current_env()
        if env is None:
            return False
        if env == 'local':
            return self._probe_lxc(env)

        return self._probe_cloud(env)

    def get_environments(self):
        if self.envs is not None:
            return self.envs
        default_home = os.path.join(os.path.expanduser('~'), '.juju')
        juju_home = os.environ.get('JUJU_HOME', default_home)

        envs_path = os.path.join(juju_home, 'environments.yaml')
        if not os.path.exists(envs_path):
            return None

        with open(envs_path) as fp:
            self.envs = yaml.safe_load(fp.read())
        return self.envs

    def get_current_env(self):
        if self.cur_env is not None:
            return self.cur_env
        envs = self.get_environments()
        self.cur_env = os.environ.get('JUJU_ENV')
        if self.cur_env is None and envs is not None:
            self.cur_env = envs.get('default', None)
        return self.cur_env

    def get_control_bucket_id(self):
        envs = self.get_environments()
        cur_env = self.get_current_env()
        try:
            return envs['environments'][cur_env]['control-bucket']
        except (KeyError, TypeError):
            # The yaml file is malformed or incomplete
            return None

    def get_provider_state(self):
        container_id = self.get_control_bucket_id()
        conn = self.swift.get_connection()
        try:
            # get_object return a tuple of (metadata, data)
            provider_state = conn.get_object(container_id, 'provider-state')[1]
            return yaml.safe_load(provider_state)
        except swiftclient.exceptions.ClientException:
            return None

    def get_nova_id(self, provider_state):
        try:
            return provider_state['state-instances'][0]
        except (TypeError, KeyError, IndexError):
            # The yaml file is malformed or incomplete
            return None

    def get_nova_instance(self, nova_id):
        try:
            return self.nova.get_client().servers.get(nova_id)
        except nova_exceptions.NotFound:
            return None

    def get_nova_ip(self, nova_instance):
        networks = nova_instance.networks.values()
        if not networks:
            return None
        # We use a single network and don't care how it's named
        net = networks[0]
        if not net:
            return None
        # Either we have a single address or the last one has been
        # added to make the instance reachable (floating or public).
        return net[-1]

    def is_ip_private(self, addr):
        """Check whether an IP address can be routed.

        :param addr: The IP address as a string.

        :returns: True if the address is private (can't be routed). False
            otherwise.
        """
        if addr.startswith('10.'):  # 10.0.0.0 - 10.255.255.255
            return True
        elif addr.startswith('192.168'):  # 192.168.0.0 - 192.168.255.255
            return True
        elif addr.startswith('172.'):  # 172.16.0.0 - 172.31.255.255
            a, b, c, d = addr.split('.')
            if 16 <= int(b) <= 31:
                return True
        return False

    def get_pid_of_sshuttle(self, ip_addr):
        try:
            cmd = ['pgrep', '-f', 'sshuttle.*%s' % ip_addr]
            pid = subprocess.check_output(cmd)
            # pgrep returns a list of pids matching the search argument
            pid.split('\n')[0]
            return int(pid)
        except subprocess.CalledProcessError:
            return None


# The single instance shared by all tests
bootstrapped_juju = BootstrappedJuju()
